"""Module for getting the Earth magnetic field"""

from __future__ import annotations

from spinifex.geometry.get_ipp import IPP, get_ipp_from_altaz, get_ipp_from_skycoord

__all__ = ["IPP", "get_ipp_from_altaz", "get_ipp_from_skycoord"]
